/*
    Copyright 2005,2006,2007,2008,2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/stat.h>
#include <openssl/aes.h>
#include <openssl/sha.h>
#include <openssl/hmac.h>

#ifdef WIN32
    #include <windows.h>
#else
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.2.4"
#define NOKEY       "NoMachineSpecificPassphraseAvailable"



#ifdef WIN32
    u8 *get_file(u8 *folder);
    int regkey(HKEY hKey, LPCTSTR lpSubKey, LPTSTR lpValueName, u8 *buff, DWORD len);
#endif
void get_key(u8 *keystr);
int hex2byte(u8 *input, u8 *output);
void AESPHM_GenerateIvFromSeed(u8 *ivSeed, u8 *iv);
int AESPHM_Decrypt(u8 *passphrase, u8 *output, u8 *input);
void steampwd(u8 *fname, u8 *keystr);
int ConnectCache_dec(void);
int getxx(u8 *data, u32 *ret, int bits);
int putxx(u8 *data, u32 num, int bits);
u8 *fd_read(u8 *name, int *fdlen);
u8 *find_data(u8 *buff, int buffsz, u8 *str);
void wait_exit(int ret);
void std_err(void);



u8      *findpass   = "Phrase",
        *dumpfile   = NULL;



int main(int argc, char *argv[]) {
#ifdef WIN32
    u8      steampath[MAX_PATH + 1];
#endif
    int     argi,
            i;
    u8      keystr[1000],
            *fname  = NULL,
            *p;

    setbuf(stdout, NULL);

    fputs("\n"
        "Steam password decoder "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    fprintf(stderr, "\n"
        "Usage: %s [options] [ClientRegistry.blob/password] [key]\n"
        "\n"
        "Options:\n"
        "-k KEY   specify the decryption key (default: calculated automatically)\n"
        "-f NAME  decrypt the data contained in the field NAME (default: \"%s\")\n"
        "-d FILE  dump the whole decrypted raw data into FILE, useful only for those\n"
        "         fields containing non-textual data like Key or AccountRecordData\n"
        "\n", argv[0], findpass);

    for(i = 1; i < argc; i++) {
        if(argv[i][0] != '-') break;
        if(((argv[i][0] != '-') && (argv[i][0] != '/')) || (strlen(argv[i]) != 2)) {
            printf("\nError: wrong argument (%s)\n", argv[i]);
            exit(1);
        }
        switch(argv[i][1]) {
            case 'k': sprintf(keystr, "%.*s", sizeof(keystr), argv[++i]);   break;
            case 'f': findpass = argv[++i];                                 break;
            case 'd': dumpfile = argv[++i];                                 break;
            default: {
                printf("\nError: wrong argument (%s)\n", argv[i]);
                exit(1);
            }
        }
    }
    argi = i;

    ConnectCache_dec();

    if(argc < (argi + 1)) {
#ifdef WIN32
        if(GetWindowLong(GetForegroundWindow(), GWL_WNDPROC)) {
            regkey(
                HKEY_CURRENT_USER,
                "Software\\Valve\\Steam", "SteamPath",
                steampath, sizeof(steampath));
            if(steampath[0]) {
                for(p = steampath; *p; p++) {
                    if(*p == '/') *p = '\\';
                }
            }
            fname = get_file(steampath);
        } else
#endif
        {
            fprintf(stderr, 
                "Error: specify the location of the blob file or directly the encrypted password\n");
            wait_exit(1);
        }
    } else {
        fname = argv[argi];
    }

    if(argc > (argi + 1)) {
        sprintf(keystr, "%.*s", sizeof(keystr), argv[argi + 1]);
    } else {
        get_key(keystr);
    }

    steampwd(fname, keystr);

    fprintf(stderr, "\n- done\n");
    wait_exit(0);
    return(0);
}



#ifdef WIN32
u8 *get_file(u8 *folder) {
    OPENFILENAME    ofn;
    static u8       filename[MAX_PATH + 1];
    static const u8 filter[] =
                    "BLOB file\0"   "*.blob\0"
                    "(*.*)\0"       "*.*\0"
                    "\0"            "\0";

    filename[0] = 0;
    memset(&ofn, 0, sizeof(ofn));
    ofn.lStructSize     = sizeof(ofn);
    ofn.lpstrFilter     = filter;
    ofn.nFilterIndex    = 1;
    ofn.lpstrFile       = filename;
    ofn.nMaxFile        = sizeof(filename);
    ofn.lpstrTitle      = "Select ClientRegistry.blob";
    ofn.lpstrInitialDir = folder;
    ofn.Flags           = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_LONGNAMES | OFN_EXPLORER;

    if(!GetOpenFileName(&ofn)) wait_exit(1);
    return(filename);
}

int regkey(HKEY hKey, LPCTSTR lpSubKey, LPTSTR lpValueName, u8 *buff, DWORD len) {
    HKEY    key;
    DWORD   type;

    buff[0] = 0;
    if(RegOpenKeyEx(hKey, lpSubKey, 0, KEY_ALL_ACCESS, &key) != ERROR_SUCCESS) {
        return(-1);
    }
    if(RegQueryValueEx(key, lpValueName, NULL, &type, buff, &len) != ERROR_SUCCESS) {
        RegCloseKey(key);
        return(-2);
    }
    RegCloseKey(key);
    if(type != REG_SZ) return(-3);
    return(len);
}

int regenum(HKEY hKey, LPCTSTR lpSubKey, u8 *vbuff, DWORD vlen, u8 *buff, DWORD len, int k) {
    HKEY    key;
    DWORD   type;

    if(RegOpenKeyEx(hKey, lpSubKey, 0, KEY_ALL_ACCESS, &key) != ERROR_SUCCESS) {
        return(-1);
    }
    if(RegEnumValue(key, k, vbuff, &vlen, NULL, &type, buff, &len) != ERROR_SUCCESS) {
        RegCloseKey(key);
        return(-2);
    }
    RegCloseKey(key);
    return(len);
}
#endif



void get_key(u8 *keystr) {
#ifdef WIN32
    u8      ProductId[256],     // 24
            MachineGuid[256],   // 37
            io[256];            // 16

    if(regkey(
        HKEY_LOCAL_MACHINE,
        "SOFTWARE\\Microsoft\\Windows\\CurrentVersion", "ProductId",
        ProductId, sizeof(ProductId)) < 0) ProductId[0] = 0;
    fprintf(stderr, "- ProductId   %s\n", ProductId);

    if(regkey(
        HKEY_LOCAL_MACHINE,
        "SOFTWARE\\Microsoft\\Cryptography", "MachineGuid",
        MachineGuid, sizeof(MachineGuid)) < 0) MachineGuid[0] = 0;
    fprintf(stderr, "- MachineGuid %s\n", MachineGuid);

    if(regkey(
        HKEY_CURRENT_USER,
        "Software\\Valve\\Half-Life\\Settings", "io",
        io, sizeof(io)) < 0) io[0] = 0;
    fprintf(stderr, "- HalfLife io %s\n", io);

    sprintf(keystr, "%s%s%s", ProductId, MachineGuid, io);
    if(!keystr[0]) strcpy(keystr, NOKEY);
    printf("\n");
#else
    fprintf(stderr, "\nError: you must provide a valid decryption key (this is not Windows!)\n");
    wait_exit(1);
#endif
}



int hex2byte(u8 *input, u8 *output) {
    static const u8 hextable[256] =
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x00\x00\x00\x00\x00\x00"
            "\x00\x0a\x0b\x0c\x0d\x0e\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x0a\x0b\x0c\x0d\x0e\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00";
    u8      *i,
            *o;

    for(i = input, o = output; i[0] && i[1]; i += 2, o++) {
        *o = (hextable[i[0]] << 4) | hextable[i[1]];
    }
    return(o - output);
}



enum {
    MINIMUM_CIPHERTEXT_LENGTH = 46,
    IV_SEED_SIZE = 8,
    IV_SIZE = 16,
    HMAC_SHA_DIGESTSIZE = 20
};



void AESPHM_GenerateIvFromSeed(u8 *ivSeed, u8 *iv) {
    int     i;
    u8      ivHash[SHA256_DIGEST_LENGTH];

    SHA256(ivSeed, IV_SEED_SIZE, ivHash);
    for(i = 0; i < 16; i++) {
        iv[i] = ivHash[i] ^ ivHash[i + 16];
    }
}



void myaesdec(AES_KEY *ctx, int length, int *ivoff, u8 *iv, u8 *input, u8 *output) {
    int     i,
            c,
            n;

    n = *ivoff;
    for(i = 0; i < length; i++) {
        if(!n) AES_ecb_encrypt(iv, iv, ctx, AES_ENCRYPT);

        c = input[i];
        output[i] = c ^ iv[n];
        iv[n] = c;
        n = (n + 1) & 15;
    }
    *ivoff = n;
}



    // thanx to an anonymous: all the encryption of Steam was just a simple "AESPHM_Decrypt(key, password, password);"
int AESPHM_Decrypt(u8 *passphrase, u8 *output, u8 *input) {
    AES_KEY ctx;
    int     inputlen,
            ivoff,
            paddingLen,
            payloadLen;
    u8      key[SHA256_DIGEST_LENGTH],
            iv[IV_SIZE],
            checkDigest[HMAC_SHA_DIGESTSIZE],
            firstPadByte,
            *digest,
            *ivSeed,
            *payload;

    input = strdup(input);

    inputlen = hex2byte(input, input);
    if(inputlen < MINIMUM_CIPHERTEXT_LENGTH) {
        fprintf(stderr, "\nError: the encrypted hex password (%d) is shorter than %d bytes\n", inputlen, MINIMUM_CIPHERTEXT_LENGTH);
        free(input);
        return(-1);
    }

	digest = input + inputlen - HMAC_SHA_DIGESTSIZE;
	ivSeed = digest - IV_SEED_SIZE;

    SHA256(passphrase, strlen(passphrase), key);
    HMAC(EVP_sha1(), key, SHA256_DIGEST_LENGTH, input, digest - input, checkDigest, NULL);

    if(memcmp(checkDigest, digest, HMAC_SHA_DIGESTSIZE)) {
        fprintf(stderr, "\nError: the hmac checksum differs, the key is wrong\n");
        free(input);
        return(-1);
    }

	AESPHM_GenerateIvFromSeed(ivSeed, iv);

    AES_set_encrypt_key(key, SHA256_DIGEST_LENGTH << 3, &ctx);
    ivoff = 0;  // AES_cfb128_encrypt in OpenSSL, but due to a difference with Crypto++ I use myaesdec
    myaesdec(&ctx, 1, &ivoff, iv, input, &firstPadByte);

    paddingLen = (firstPadByte & 15) + 3;
    if((input + paddingLen) > ivSeed) {
        fprintf(stderr, "\nError: the initial padded bytes (%d) are longer than the seed (%d)\n", paddingLen, IV_SEED_SIZE);
        free(input);
        return(-1);
    }

    payload = input + paddingLen;
    payloadLen = ivSeed - payload;
    if(payloadLen) {
        myaesdec(&ctx, paddingLen - 1, &ivoff, iv, input + 1, output);
        myaesdec(&ctx, payloadLen, &ivoff, iv, payload, output);
    }

    free(input);
    return(payloadLen);
}



void do_dumpfile(u8 *data, int len) {
    static int  accesses = 0;
    FILE    *fd_dump;

    if(!dumpfile) return;
    fd_dump = fopen(dumpfile, accesses ? "ab" : "wb");
    if(!fd_dump) return;
    fwrite(data, 1, len, fd_dump);
    fclose(fd_dump);
    accesses++;
}



void steampwd(u8 *fname, u8 *keystr) {
    int     len,
            nlen,
            fdsize,
            pwds;
    u8      *fdbuff,
            *fdnext,
            *p;

    p = malloc(strlen(findpass) + 2 + 1);
    sprintf(p, "%s\x01\x50", findpass);
    findpass = p;

    fprintf(stderr, "- decryption key:\n  \"%s\"\n\n", keystr);

    fdbuff = fd_read(fname, &fdsize);
    if(!fdbuff) {
        fprintf(stderr, "- the input will be considered a password\n");
        len = strlen(fname);
        fdsize = 64 + len;
        fdbuff = malloc(fdsize);
        p = fdbuff;
        p += sprintf(p, "%-30s", findpass);
        p += putxx(p, 0,    16);
        p += putxx(p, len,  32);
        strcpy(p, fname);
    }

    printf("\n");

    fdnext = fdbuff;
    for(pwds = 0;; pwds++) {    // totally useless, steam saves only the password of the last logged user
        p = find_data(fdnext, fdsize, findpass);
        if(!p) {
            if(pwds) break;
            fprintf(stderr, "\n"
                "- password not found, the possible reasons are that no account has been used\n"
                "  yet or it's a new version of Steam (like a beta) which uses a new way to\n"
                "  store the saved password or, more probably, a hash-like string.\n"
                "\n"
                "  remember that in any case Steam (any version till the current one in the\n"
                "  moment I'm writing) saves the last valid password even if both the saving of\n"
                "  the password and the storing of the local account informations are disabled\n");
            wait_exit(1);
        }

        // the blob file format is not so hard but I want the fastest solution
        p = strchr(p, 0x01);
        p += 0x18;
        p += getxx(p, &nlen,    16);
        p += getxx(p, &len,     32);
        p += nlen;

        fdsize -= (p - fdnext);
        fdnext = p;

        if((len > 0) && (len < fdsize)) {
            p[len] = 0;
            len = AESPHM_Decrypt(keystr, p, p);     // try the retrieved key
            if((len < 0) && strcmp(keystr, NOKEY)) {
                fprintf(stderr, "- now I try with the decryption key \"%s\"\n", NOKEY);
                len = AESPHM_Decrypt(NOKEY, p, p);  // if the first fails try with NOKEY
            }
            if(len >= 0) {
                printf("  PASSWORD: %.*s\n", len, p);
                do_dumpfile(p, len);
            }
        }
    }
    free(fdbuff);
}



static u32  steam_seed1 = 0,
            steam_seed2 = 0,
            steam_seed3[] = {
                0x1739a3b0, 0xb8907fe1, 0x8290d3b7, 0x72839cd0,
                0x242df096, 0x3829750b, 0x38de7a77, 0x72f0924c,
                0x44783927, 0x01925372, 0x20902714, 0x27585920,
                0x27890632, 0x82910476, 0x72906721, 0x28798904
            };

void steam_srand(u32 init) {
    steam_seed1 = 0;    // CDQ of init
    steam_seed2 = init;
}

u32 steam_rand(void) {
    steam_seed2++;
    steam_seed1 += 0x25fe6761;
    return(steam_seed3[steam_seed2 % 17] ^ steam_seed1);
}

int steam_rand_crypt(u32 init, u8 *data, int len) {
    u32     *p32,
            crc;
    int     i;

    steam_srand(init);

    p32 = (u32 *)data;
    len /= 4;
    len--;  // last 32bit is crc
    crc = 0;
    for(i = 0; i < len; i++) {
        p32[i] ^= steam_rand();
        crc    += p32[i];
    }
    crc ^= steam_rand();
    if(crc != p32[i]) return(-1);
    return(0);
}



int ConnectCache_dec(void) {
#ifdef WIN32
    u32     mtbf;
    int     i,
            len;
    u8      value[256],
            buff[1024];

    len = regkey(
        HKEY_LOCAL_MACHINE,
        "SOFTWARE\\Valve\\Steam", "MTBF",
        buff, sizeof(buff));
    if((len < 0) && (len != -3)) return(-1);
    getxx(buff, &mtbf, 32);

    fprintf(stderr, "\n- ConnectCache cookies with MTBF 0x%08x (NOT the passwords):\n\n", mtbf);

    for(i = 0;; i++) {
        len = regenum(
            HKEY_LOCAL_MACHINE,
            "SOFTWARE\\Valve\\Steam\\ConnectCache", // %x%x, it's the crc32 of the username plus 1
            value, sizeof(value),
            buff, sizeof(buff),
            i);
        if(len < 0) break;

        len -= 4;   // skip the type value which must be != 1
        if(len > 0) {
            memmove(buff, buff + 4, len);
            if(!steam_rand_crypt(mtbf, buff, len)) {
                printf("  %s: %.*s\n\n", value, len, buff);
                //do_dumpfile(buff, len);
            }
        }
    }
    return(0);
#else
    return(-1);
#endif
}



int getxx(u8 *data, u32 *ret, int bits) {
    u32     num;
    int     i,
            bytes;

    bytes = bits >> 3;
    for(num = i = 0; i < bytes; i++) {
        num |= (data[i] << (i << 3));
    }
    *ret = num;
    return(bytes);
}



int putxx(u8 *data, u32 num, int bits) {
    int     i,
            bytes;

    bytes = bits >> 3;
    for(i = 0; i < bytes; i++) {
        data[i] = (num >> (i << 3));
    }
    return(bytes);
}



u8 *fd_read(u8 *name, int *fdlen) {
    struct stat xstat;
    FILE        *fd;
    u8     *buff;

    fprintf(stderr, "- open file %s\n", name);
    fd = fopen(name, "rb");
    if(!fd) return(NULL);
    fstat(fileno(fd), &xstat);
    buff = malloc(xstat.st_size);
    if(!buff) std_err();
    fread(buff, 1, xstat.st_size, fd);
    fclose(fd);
    *fdlen = xstat.st_size;
    return(buff);
}



u8 *find_data(u8 *buff, int buffsz, u8 *str) {
    int     strsz;
    u8      *p,
            *limit;

    strsz = strlen(str);
    limit = buff + buffsz - strsz;
    for(p = buff; p <= limit; p++) {
        if(!memcmp(p, str, strsz)) return(p);
    }
    return(NULL);
}



void wait_exit(int ret) {
#ifdef WIN32
    if(GetWindowLong(GetForegroundWindow(), GWL_WNDPROC)) {
        fprintf(stderr, "\n    Press RETURN to exit\n");
        fgetc(stdin);
    }
#endif
    exit(ret);
}



void std_err(void) {
    perror("\nError");
    wait_exit(1);
}


